from __future__ import annotations

from typing import Any, List, TYPE_CHECKING

import O365.message as message
import O365.utils.utils as utils

from subtypes import Str, Markup
from pathmagic import Dir, PathLike
from iotools import HtmlGui

from .attribute import Attribute, NonFilterableAttribute, EnumerativeAttribute, BooleanAttribute
from .query import Query, BulkAction, BulkActionContext
from .fluent import FluentMessage

if TYPE_CHECKING:
    from .office import Office


class Message(message.Message):
    def __init__(self, *args: Any, office: Office = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.office = office

    def __repr__(self) -> str:
        return f"{type(self).__name__}(subject={repr(self.subject)}, from={repr(self.sender.address)}, is_read={self.is_read}, importance={repr(self.importance.value)}, attachments={len(self.attachments)}, received={self.received})"

    def __str__(self) -> str:
        return self.text

    def _repr_html_(self) -> str:
        return f"<strong><mark>{self.subject}</mark></strong><br><br>{self.body}"

    @property
    def text(self) -> str:
        return str(Str(Markup(self.body).text).re.sub(r"<!--.*?-->", "").re.sub(r"(?<=\S)(\s)*?\n(\s)*?\n(\s)*?(?=\S)", "\n\n").strip())

    @property
    def markup(self) -> Markup:
        return Markup(self.body)

    def reply(self, *args: Any, **kwargs: Any) -> FluentMessage:
        message = super().reply(*args, **kwargs)
        message.office = self.office
        return FluentMessage(message=message)

    def forward(self, *args: Any, **kwargs: Any) -> FluentMessage:
        message = super().forward(*args, **kwargs)
        message.office = self.office
        return FluentMessage(message=message)

    def copy(self, *args: Any, **kwargs: Any) -> FluentMessage:
        message = super().copy(*args, **kwargs)
        message.office = self.office
        return FluentMessage(message=message)

    def fluent(self) -> FluentMessage:
        return FluentMessage(message=self)

    def render(self) -> None:
        HtmlGui(name=self.subject, text=self.body)

    def save_attachments_to(self, path: PathLike) -> bool:
        if not self.has_attachments:
            return []
        else:
            self.attachments.download_attachments()
            for attachment in self.attachments:
                attachment.save(path)

            return [Dir(path).files[attachment.name] for attachment in self.attachments]

    class Attributes:
        class From(Attribute):
            name = "from"

        class Sender(Attribute):
            name = "sender"

        class Subject(Attribute):
            name = "subject"

        class ReceivedOn(Attribute):
            name = "received_date_time"

        class LastModified(Attribute):
            name = "last_modified_date_time"

        class Categories(Attribute):
            name = "categories"

        class IsRead(BooleanAttribute):
            name = "is_read"

        class HasAttachments(BooleanAttribute):
            name = "has_attachments"

        class IsDraft(BooleanAttribute):
            name = "is_draft"

        class HasDeliveryReceipt(BooleanAttribute):
            name = "is_delivery_receipt_requested"

        class HasReadReceipt(BooleanAttribute):
            name = "is_read_receipt_requested"

        class Importance(EnumerativeAttribute):
            name, enumeration = "importance", utils.ImportanceLevel

        class Body(NonFilterableAttribute):
            name = "body"

        class Cc(NonFilterableAttribute):
            name = "cc_recipients"

        class Bcc(NonFilterableAttribute):
            name = "bcc_recipients"

        class To(NonFilterableAttribute):
            name = "to_recipients"


class MessageQuery(Query):
    @property
    def bulk(self) -> BulkMessageAction:
        return BulkMessageAction(self)

    def execute(self) -> List[Message]:
        super().execute()

        messages = list(self._container.get_messages(limit=self._limit, query=self._query, order_by=self._order))
        for message_ in messages:
            message_.office = self._container.office

        return messages


class BulkMessageAction(BulkAction):
    def copy(self, folder: Any) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=Message.copy, args=(folder,))

    def move(self, folder: Any) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=Message.move, args=(folder,))

    def delete(self) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=Message.delete)

    def mark_as_read(self) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=Message.mark_as_read)

    def save_draft(self) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=Message.save_draft)



from __future__ import annotations

from typing import Any, List, TYPE_CHECKING

import O365.mailbox as mailbox
import O365.address_book as address_book

from miscutils import LazyProperty

from .attribute import Attribute, NonFilterableAttribute
from .message import Message, MessageQuery
from .contact import Contact, ContactQuery
from .query import Query, BulkAction, BulkActionContext
from .fluent import FluentMessage

if TYPE_CHECKING:
    from .office import Office


class Folder:
    @property
    def message(self) -> FluentMessage:
        return FluentMessage(parent=self)


class MessageFolder(mailbox.Folder, Folder):
    message_constructor = Message

    def __init__(self, *args: Any, office: Office = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.office = office

    @property
    def folders(self) -> MessageFolderQuery:
        return MessageFolderQuery(container=self)

    @property
    def messages(self) -> MessageQuery:
        return MessageQuery(container=self)

    @staticmethod
    def order_messages_by_date(messages: list, desc: bool = True) -> list:
        return sorted(messages, reverse=desc, key=lambda val: val.received)

    class Attributes:
        class ChildFolderCount(Attribute):
            name = "child_folder_count"

        class TotalItemCount(Attribute):
            name = "total_item_count"

        class UnreadItemCount(Attribute):
            name = "unread_item_count"

        class Name(Attribute):
            name = "display_name"

        class ChildFolders(NonFilterableAttribute):
            name = "child_folders"

        class Messages(NonFilterableAttribute):
            name = "messages"


class ContactFolder(address_book.ContactFolder, Folder):
    message_constructor = Message
    contact_constructor = Contact

    def __init__(self, *args: Any, office: Office = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.office = office

    @property
    def folders(self) -> ContactFolderQuery:
        return ContactFolderQuery(container=self)

    @property
    def contacts(self) -> ContactQuery:
        return ContactQuery(container=self)

    def from_address(self, address: str) -> Contact:
        contact = self.get_contact_by_email(address)
        if contact is None:
            return None
        else:
            contact.office = self.office
            return contact

    class Attributes:
        class Name(Attribute):
            name = "display_name"

        class Contacts(NonFilterableAttribute):
            name = "contacts"

        class ChildFolders(NonFilterableAttribute):
            name = "child_folders"


class Folders:
    def __init__(self, office: Office) -> None:
        self.office = office


class MessageFolders(Folders):
    def __init__(self, office: Office) -> None:
        super().__init__(office=office)
        self._mailbox = office.account.mailbox()
        self._mailbox.folder_constructor = MessageFolder

    @LazyProperty
    def main(self) -> MessageFolder:
        return MessageFolder(parent=self.office.account, main_resource=self.office.account.main_resource, name='MailBox', root=True, office=self.office)

    @LazyProperty
    def inbox(self) -> MessageFolder:
        folder = self._mailbox.inbox_folder()
        folder.office = self.office
        return folder

    @LazyProperty
    def outbox(self) -> MessageFolder:
        folder = self._mailbox.outbox_folder()
        folder.office = self.office
        return folder

    @LazyProperty
    def sent(self) -> MessageFolder:
        folder = self._mailbox.sent_folder()
        folder.office = self.office
        return folder

    @LazyProperty
    def drafts(self) -> MessageFolder:
        folder = self._mailbox.drafts_folder()
        folder.office = self.office
        return folder

    @LazyProperty
    def junk(self) -> MessageFolder:
        folder = self._mailbox.junk_folder()
        folder.office = self.office
        return folder

    @LazyProperty
    def deleted(self) -> MessageFolder:
        folder = self._mailbox.deleted_folder()
        folder.office = self.office
        return folder

    def custom(self, folder_name: str = None, folder_id: int = None) -> MessageFolder:
        folder = self._mailbox.get_folder(folder_name=folder_name, folder_id=folder_id)
        folder.office = self.office
        return folder


class ContactFolders(Folders):
    def __init__(self, office: Office) -> None:
        super().__init__(office=office)
        self._book = office.account.address_book()
        self._book.contact_constructor, self._book.message_constructor = Contact, Message

    @LazyProperty
    def main(self) -> ContactFolder:
        return ContactFolder(parent=self.office.account, main_resource=self.office.account.main_resource, name='AddressBook', root=True, office=self.office)

    @LazyProperty
    def global_(self) -> ContactFolder:
        folder = self.office.account.address_book(address_book="gal")
        folder.office = self.office
        folder.contact_constructor = Contact
        return folder


class BulkFolderAction(BulkAction):
    def move(self, folder: Folder) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=MessageFolder.move_folder, args=(folder,))

    def delete(self) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=MessageFolder.delete)


class BulkMessageFolderAction(BulkFolderAction):
    def copy(self, folder: Folder) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=MessageFolder.copy_folder, args=(folder,))


class BulkContactFolderAction(BulkFolderAction):
    pass


class FolderQuery(Query):
    def __getitem__(self, key: str) -> Folder:
        raise NotImplementedError()

    def execute(self) -> List[Message]:
        super().execute()
        folders = list(self._container.get_folders(limit=self._limit, query=self._query, order_by=self._order))
        for folder in folders:
            folder.office = self._container.office

        return folders


class MessageFolderQuery(FolderQuery):
    def __getitem__(self, key: str) -> MessageFolder:
        folder = self._container.get_folder(folder_name=key)
        folder.office = self._container.office
        return folder

    @property
    def bulk(self) -> BulkMessageFolderAction:
        return BulkMessageFolderAction(self)


class ContactFolderQuery(FolderQuery):
    def __getitem__(self, key: str) -> ContactFolder:
        return self._container.get_folder(folder_name=key)

    @property
    def bulk(self) -> BulkContactFolderAction:
        return BulkContactFolderAction(self)



from __future__ import annotations

import webbrowser

from maybe import Maybe
from pathmagic import Dir, File
from miscutils import Supressor, LazyProperty

with Supressor():
    import O365 as off

if True:
    from .config import Config
    from .contact import ContactNameSpace
    from .fluent import FluentMessage
    from .folder import MessageFolders, ContactFolders
    from office import resources


class Office:
    scopes = [
        "offline_access", "User.Read", "Mail.ReadWrite", "Mail.Send", "Mail.ReadWrite.Shared", "Mail.Send.Shared",
        "Contacts.ReadWrite", "Contacts.ReadWrite.Shared", "Calendars.ReadWrite", "Calendars.ReadWrite.Shared",
        "Files.ReadWrite", "Files.ReadWrite.All", "Sites.Read.All", "Sites.ReadWrite.All",
    ]

    def __init__(self, email_address: str = None, connection: str = None) -> None:
        self.token = off.FileSystemTokenBackend(token_path=Dir.from_home().path, token_filename="o365_token.txt")

        self.config = Config()
        self.connection = Maybe(connection).else_(self.config.data.default_connections.office)

        settings = self.config.data.connections.office[self.connection]

        self.address = Maybe(email_address).else_(settings.default_email)
        self.account = off.Account((settings.id, settings.secret), main_resource=self.address, token_backend=self.token)

        self.outlook, self.people = Outlook(self), People(self)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(account={self.address})"

    def request_token(self) -> None:
        auth_url = self.account.connection.get_authorization_url(requested_scopes=self.scopes)
        webbrowser.open(auth_url)
        self.account.connection.request_token(input("Please follow the link that will open momentarily and grant permission. Then enter the url of the inbox page you land on.\n\n"))


class Manager:
    def __init__(self, office: Office) -> None:
        self.office = office


class Outlook(Manager):
    def __init__(self, office: Office) -> None:
        super().__init__(office=office)
        self._signature = self.office.config.appdata.newfile("signature", "html")

    @LazyProperty
    def folders(self) -> MessageFolders:
        return MessageFolders(self.office)

    @property
    def signature(self) -> str:
        return self._signature.contents

    @signature.setter
    def signature(self, signature: str) -> None:
        self._signature.contents = signature

    @property
    def message(self) -> FluentMessage:
        return FluentMessage(parent=self.folders.main)


class People(Manager):
    def __init__(self, office: Office) -> None:
        super().__init__(office=office)
        self.contacts = ContactNameSpace(self.office)

    @LazyProperty
    def folders(self) -> ContactFolders:
        return ContactFolders(self.office)


class BlobStorage:
    def __init__(self, connection: str = None) -> None:
        self.blob_type_mappings = File.from_resource(package=resources, name="blob_content_types", extension="json").contents

        self.config = Config()
        self.connection = Maybe(connection).else_(self.config.data.default_connection)

        self.authenticate()

    def authenticate(self) -> None:
        import azure.storage.blob as blob
        from .blob import BlobContainerNameSpace

        settings = self.config.data.connections.blob[self.connection]
        self.blob, self.service = blob, blob.BlockBlobService(account_name=settings.account, account_key=settings.key)
        self.containers = BlobContainerNameSpace(self)



import pytest


def test_BulkContactAction_delete():
    assert True


def test_ContactNameSpace___init__():
    assert True


def test_ContactQuery_execute():
    assert True


def test_Contact___init__():
    assert True


def test_Contact___repr__():
    assert True



import pytest


def test_BulkActionContext___bool__():
    assert True


def test_BulkActionContext___enter__():
    assert True


def test_BulkActionContext___exit__():
    assert True


def test_BulkActionContext___init__():
    assert True


def test_BulkActionContext___len__():
    assert True


def test_BulkActionContext__execute_query():
    assert True


def test_BulkActionContext__perform_bulk_action():
    assert True


def test_BulkActionContext_commit():
    assert True


def test_BulkActionContext_execute():
    assert True


def test_BulkAction___init__():
    assert True


def test_Query___call__():
    assert True


def test_Query___init__():
    assert True


def test_Query__build_clause_chunk():
    assert True


def test_Query__build_select_clause():
    assert True


def test_Query__build_where_clause():
    assert True


def test_Query__negation_context():
    assert True


def test_Query_execute():
    assert True


def test_Query_get_messages():
    assert True


def test_Query_limit():
    assert True


def test_Query_order_by():
    assert True


def test_Query_select():
    assert True


def test_Query_where():
    assert True



from __future__ import annotations

from typing import Any, Callable, List, NoReturn, Optional, Tuple

import O365.utils.utils as utils
from subtypes import Enum


class BaseAttributeMeta(type):
    name: str

    def __eq__(self, other: Any) -> BooleanExpression:  # type: ignore
        return BooleanExpression(self.name, utils.Query.equals, other)

    def __ne__(self, other: Any) -> BooleanExpression:  # type: ignore
        return BooleanExpression(self.name, utils.Query.unequal, other)

    def __lt__(self, other: Any) -> BooleanExpression:
        return BooleanExpression(self.name, utils.Query.less, other)

    def __le__(self, other: Any) -> BooleanExpression:
        return BooleanExpression(self.name, utils.Query.less_equal, other)

    def __gt__(self, other: Any) -> BooleanExpression:
        return BooleanExpression(self.name, utils.Query.greater, other)

    def __ge__(self, other: Any) -> BooleanExpression:
        return BooleanExpression(self.name, utils.Query.greater_equal, other)

    def __and__(self, other: Any) -> BooleanExpressionClause:
        return self._resolve() & other._resolve()

    def __or__(self, other: Any) -> BooleanExpressionClause:
        return self._resolve() | other._resolve()

    def _resolve(self) -> Any:
        return BooleanExpression(self.name)._resolve()


class BooleanAttributeMeta(BaseAttributeMeta):
    def __invert__(self) -> BooleanExpression:
        return BooleanExpression(self.name, utils.Query.equals, True).negate()

    def _resolve(self) -> BooleanExpressionClause:
        return BooleanExpression(self.name, utils.Query.equals, True)._resolve()


class EnumerativeAttributeMeta(BaseAttributeMeta):
    def __new__(mcs, classname: str, bases: tuple, attributes: dict) -> Any:
        def make_function(func_name: str, value: str) -> Callable:
            def template(cls: EnumerativeAttributeMeta) -> BooleanExpression:
                return BooleanExpression(cls.name, utils.Query.equals, value)

            template.__name__ = func_name
            return template

        if attributes.get("enumeration") is not None:
            funcnames = [(f"is_{enum.name.lower()}", enum.value) for enum in attributes["enumeration"]]
            for name, val in funcnames:
                attributes[name] = classmethod(make_function(func_name=name, value=val))

        return type.__new__(mcs, classname, bases, attributes)


class NonFilterableMeta(type):
    def __getattr__(self, attr: str) -> NoReturn:
        raise AttributeError("This attribute cannot be used in the filter/where clause of a query.")


class BaseAttribute:
    name: str = None

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"


class NonFilterableAttribute(BaseAttribute, metaclass=NonFilterableMeta):
    pass


class FilterableAttribute(BaseAttribute):
    def __init__(self, order_by: str):
        self.order_by = order_by

    @classmethod
    def contains(cls, item: Any) -> BooleanExpression:
        return BooleanExpression(cls.name, utils.Query.contains, item)

    @classmethod
    def startswith(cls, text: str) -> BooleanExpression:
        return BooleanExpression(cls.name, utils.Query.startswith, text)

    @classmethod
    def endswith(cls, text: str) -> BooleanExpression:
        return BooleanExpression(cls.name, utils.Query.endswith, text)

    @classmethod
    def asc(cls) -> FilterableAttribute:
        return cls(Direction.ASCENDING)

    @classmethod
    def desc(cls) -> FilterableAttribute:
        return cls(Direction.DESCENDING)


class Attribute(FilterableAttribute, metaclass=BaseAttributeMeta):
    pass


class BooleanAttribute(FilterableAttribute, metaclass=BooleanAttributeMeta):
    pass


class EnumerativeAttribute(FilterableAttribute, metaclass=EnumerativeAttributeMeta):
    enumeration = None


class BooleanExpression:
    logical_opposites = {
        utils.Query.equals: utils.Query.unequal,
        utils.Query.unequal: utils.Query.equals,
        utils.Query.greater: utils.Query.less_equal,
        utils.Query.less_equal: utils.Query.greater,
        utils.Query.less: utils.Query.greater_equal,
        utils.Query.greater_equal: utils.Query.less,
    }

    def __init__(self, attribute_name: str, query_func: Callable = None, argument: Any = None) -> None:
        self.func, self.attr, self.arg, self.negated = query_func, attribute_name, argument, False

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join([f'{attr}={repr(val)}' for attr, val in self.__dict__.items() if not attr.startswith('_')])})"

    def __and__(self, other: Any) -> BooleanExpressionClause:
        return self._resolve() & other._resolve()

    def __or__(self, other: Any) -> BooleanExpressionClause:
        return self._resolve() | other._resolve()

    def __invert__(self) -> BooleanExpression:
        return self.negate()

    def negate(self) -> BooleanExpression:
        if self.func in self.logical_opposites:
            self.func = self.logical_opposites[self.func]
        else:
            self.negated = not self.negated

        return self

    def _resolve(self) -> BooleanExpressionClause:
        return BooleanExpressionClause(self)


class BooleanExpressionClause:
    def __init__(self, expression: BooleanExpression) -> None:
        self.expressions: List[Tuple[BooleanExpression, Optional[utils.ChainOperator]]] = [(expression, None)]

    def __and__(self, other: Any) -> BooleanExpressionClause:
        return self._coalesce(other._resolve(), utils.ChainOperator.AND)

    def __or__(self, other: Any) -> BooleanExpressionClause:
        return self._coalesce(other._resolve(), utils.ChainOperator.OR)

    def _coalesce(self, clause: BooleanExpressionClause, operator: utils.ChainOperator) -> BooleanExpressionClause:
        clause.expressions[0] = (clause.expressions[0][0], operator)
        self.expressions.extend(clause.expressions)
        return self

    def _resolve(self) -> BooleanExpressionClause:
        return self


class Direction(Enum):
    ASCENDING, DESCENDING = "asc", "desc"



__all__ = ["Office", "BlobStorage", "MessageFolder", "ContactFolder", "Message", "Contact", "Config"]

from .office import Office, BlobStorage
from .folder import MessageFolder, ContactFolder
from .message import Message
from .contact import Contact
from .config import Config



import pytest


def test_FluentMessage___init__():
    assert True


def test_FluentMessage_attach():
    assert True


def test_FluentMessage_body():
    assert True


def test_FluentMessage_from_():
    assert True


def test_FluentMessage_send():
    assert True


def test_FluentMessage_sign():
    assert True


def test_FluentMessage_subject():
    assert True


def test_FluentMessage_to():
    assert True



from __future__ import annotations

import os
from typing import Any, TYPE_CHECKING

from pathmagic import File, PathLike
from miscutils import NameSpace

if TYPE_CHECKING:
    from .office import BlobStorage


class BlobContainerNameSpace(NameSpace):
    def __init__(self, blob_manager: BlobStorage) -> None:
        self._manager = blob_manager
        super().__init__({container.name: BlobContainer(container=container, blob_manager=self._manager) for container in self._manager.service.list_containers()})


class BlobContainer:
    def __init__(self, container: Any, blob_manager: BlobStorage) -> None:
        self.raw, self.name, self.manager = container, container.name, blob_manager
        self.service = self.manager.service
        self._cached_len = 0

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={repr(self.name)}, num_blobs={self._cached_len or '?'})"

    def __str__(self) -> str:
        return self.name

    def __len__(self) -> int:
        self._cached_len = sum([1 for blob in self.service.list_blob_names(self.name)])
        return self._cached_len

    def __bool__(self) -> bool:
        return len(self) > 0

    def __iter__(self) -> BlobContainer:
        self.__iter = iter(self.service.list_blobs(self.name))
        return self

    def __next__(self) -> Any:
        return Blob(blob=next(self.__iter), container=self)

    def __getitem__(self, key: str) -> Blob:
        return Blob(blob=self.service.get_blob_properties(container_name=self.name, blob_name=key), container=self)

    def download_blob_to(self, blob_name: str, path: PathLike) -> PathLike:
        self[blob_name].download_to(path)
        return File.from_pathlike(path)

    def upload_blob_from(self, blob_name: str, path: PathLike) -> Blob:
        from azure.storage.blob.models import ContentSettings

        file = File(path)
        content_type = self.manager.blob_type_mappings[file.extension]
        self.service.create_blob_from_path(container_name=self.name, blob_name=blob_name, file_path=str(file), content_settings=ContentSettings(content_type=content_type))

        return self[blob_name]


class Blob:
    def __init__(self, blob: Any, container: BlobContainer) -> None:
        self.raw, self.name, self.container = blob, blob.name, container
        self.service = self.container.manager.service

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={self.name}, container={self.container.name})"

    def download_to(self, path: PathLike) -> PathLike:
        self.service.get_blob_to_path(container_name=self.container.name, blob_name=self.name, file_path=os.fspath(path))
        return File.from_pathlike(path)

    def delete(self) -> None:
        self.service.delete_blob(container_name=self.container.name, blob_name=self.name)



import pytest


def test_BulkContactFolderAction_delete():
    assert True


def test_BulkContactFolderAction_move():
    assert True


def test_BulkFolderAction_delete():
    assert True


def test_BulkFolderAction_move():
    assert True


def test_BulkMessageFolderAction_copy():
    assert True


def test_BulkMessageFolderAction_delete():
    assert True


def test_BulkMessageFolderAction_move():
    assert True


def test_ContactFolderQuery___getitem__():
    assert True


def test_ContactFolderQuery_execute():
    assert True


def test_ContactFolder___init__():
    assert True


def test_ContactFolders___init__():
    assert True


def test_FolderQuery___getitem__():
    assert True


def test_FolderQuery_execute():
    assert True


def test_Folders___init__():
    assert True


def test_MessageFolderQuery___getitem__():
    assert True


def test_MessageFolderQuery_execute():
    assert True


def test_MessageFolder___init__():
    assert True


def test_MessageFolder_order_messages_by_date():
    assert True


def test_MessageFolders___init__():
    assert True


def test_MessageFolders_custom():
    assert True



import pytest


def test_BulkMessageAction_copy():
    assert True


def test_BulkMessageAction_delete():
    assert True


def test_BulkMessageAction_mark_as_read():
    assert True


def test_BulkMessageAction_move():
    assert True


def test_BulkMessageAction_save_draft():
    assert True


def test_MessageQuery_execute():
    assert True


def test_Message___init__():
    assert True


def test_Message___repr__():
    assert True


def test_Message___str__():
    assert True


def test_Message__repr_html_():
    assert True


def test_Message_copy():
    assert True


def test_Message_fluent():
    assert True


def test_Message_forward():
    assert True


def test_Message_render():
    assert True


def test_Message_reply():
    assert True


def test_Message_save_attachments_to():
    assert True



from __future__ import annotations

from typing import List, Dict, Any, TYPE_CHECKING

import O365.address_book as address_book
from subtypes import Str
from miscutils import NameSpace

from .attribute import Attribute, NonFilterableAttribute
from .message import Message
from .query import Query, BulkAction, BulkActionContext
from .fluent import FluentMessage

if TYPE_CHECKING:
    from .office import Office


class Contact(address_book.Contact):
    message_constructor = Message

    def __init__(self, *args: Any, office: Office = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.office = office

    def __repr__(self) -> str:
        return f"{type(self).__name__}(name={repr(self.full_name)}, email={repr(self.main_email)})"

    def __str__(self) -> str:
        return self.name

    @property
    def message(self) -> FluentMessage:
        message = self.new_message()
        message.office = self.office
        return FluentMessage(message=message)

    class Attributes:
        class Name(Attribute):
            name = "given_name"

        class Surname(Attribute):
            name = "surname"

        class DisplayName(Attribute):
            name = "display_name"

        class Company(Attribute):
            name = "company_name"

        class Department(Attribute):
            name = "department"

        class Created(Attribute):
            name = "created_date_time"

        class LastModified(Attribute):
            name = "last_modified_date_time"

        class HomeAddress(Attribute):
            name = "home_address"

        class JobTitle(Attribute):
            name = "job_title"

        class Manager(Attribute):
            name = "manager"

        class MiddleName(Attribute):
            name = "middle_name"

        class Mobile(Attribute):
            name = "mobile_phone1"

        class OfficeLocation(Attribute):
            name = "office_location"

        class Profession(Attribute):
            name = "profession"

        class EmailAddresses(NonFilterableAttribute):
            name = "email_addresses"


class ContactNameSpace(NameSpace):
    def __init__(self, office: Office) -> None:
        self._office = office

        book = self._office.account.address_book()
        book.contact_constructor = Contact

        contacts_by_name: Dict[str, List[Contact]] = {}
        for contact in book.get_contacts():
            contacts_by_name.setdefault(Str(contact.name).case.snake(), []).append(contact)
            contacts_by_name.setdefault(Str(contact.display_name).case.snake(), []).append(contact)

        mappings = {}
        for name, contacts in contacts_by_name.items():
            if len(contacts) == 1 and name and name.lower() != "none" and not hasattr(self, name):
                contacts[0].office = self._office
                mappings[name] = contacts[0]

        super().__init__(mappings)


class ContactQuery(Query):
    @property
    def bulk(self) -> BulkContactAction:
        return BulkContactAction(self)

    def execute(self) -> List[Contact]:
        super().execute()
        result = self._container.get_contacts(limit=self._limit, query=self._query, order_by=self._order)

        for contact in result:
            contact.office = self._container.office

        return result


class BulkContactAction(BulkAction):
    def delete(self) -> BulkActionContext:
        return BulkActionContext(query=self._query, action=Contact.delete)



from __future__ import annotations

import contextlib
from typing import Any, Callable, Collection, Generator, Tuple

import O365.utils.utils as utils

from maybe import Maybe

from .attribute import BaseAttribute, Attribute, FilterableAttribute, BooleanExpression, BooleanExpressionClause


class Query:
    def __init__(self, container: Any) -> None:
        self._container = container
        self._casing_function = self._container.protocol.casing_function
        self._query = utils.Query(protocol=self._container.protocol)
        self._select: Tuple[BaseAttribute, ...] = None
        self._where: BooleanExpressionClause = None
        self._order: str = None
        self._limit: int = None

    def __call__(self) -> Any:
        return self.execute()

    @property
    def bulk(self) -> BulkAction:
        return BulkAction(self)

    def select(self, *args: Tuple[BaseAttribute, ...]) -> Query:
        self._select = args
        return self

    def where(self, *args: Any) -> Query:
        item, = args
        self._where = item._resolve()
        return self

    def order_by(self, order_clause: Any) -> Query:
        if isinstance(order_clause, str):
            self._order = order_clause
        elif isinstance(order_clause, Attribute) and issubclass(order_clause, FilterableAttribute):
            attribute = order_clause.asc()
            self._order = f"{self._casing_function(attribute.name)} {attribute.order_by}"
        elif isinstance(order_clause, FilterableAttribute):
            self._order = f"{self._casing_function(order_clause.name)} {order_clause.order_by}"
        else:
            raise TypeError(f"Unrecognized type '{type(order_clause)}' of '{order_clause}' for 'order_by'.")

        return self

    def limit(self, limit: int = 25) -> Query:
        self._limit = limit
        return self

    def execute(self) -> Any:
        self._build_select_clause()
        self._build_where_clause()

    def get_messages(self) -> list:
        return self.execute()

    def _build_select_clause(self) -> None:
        if self._select:
            self._query.select(*[self._casing_function(attribute.name) for attribute in self._select])

    def _build_where_clause(self) -> None:
        if self._where is not None:
            for expression, operator in self._where.expressions:
                self._build_clause_chunk(expression, operator)

    def _build_clause_chunk(self, expression: BooleanExpression, operator: utils.ChainOperator) -> None:
        if operator is not None:
            self._query = self._query.chain(operator.value)

        self._query = self._query.on_attribute(self._casing_function(expression.attr))

        with self._negation_context() if expression.negated else contextlib.nullcontext():
            self._query = expression.func(self._query, expression.arg)

    @contextlib.contextmanager
    def _negation_context(self) -> Generator[None, None, None]:
        self._query = self._query.negate()

        try:
            yield None
        finally:
            self.query = self._query.negate()


class BulkActionContext:
    def __init__(self, query: Query, action: Callable, args: Any = None, kwargs: Any = None) -> None:
        self._query, self._action, self._args, self._kwargs, self._committed = query, action, Maybe(args).else_(()), Maybe(kwargs).else_({}), False
        self.result: Collection = []

    def __len__(self) -> int:
        return len(self.result)

    def __bool__(self) -> bool:
        return len(self) > 0

    def commit(self) -> None:
        self._committed = True

    def execute(self) -> int:
        self._execute_query()
        self._perform_bulk_action()
        return len(self)

    def __enter__(self) -> BulkActionContext:
        self._execute_query()
        return self

    def __exit__(self, ex_type: Any, ex_value: Any, ex_traceback: Any) -> None:
        if self._committed:
            self._perform_bulk_action()

    def _execute_query(self) -> None:
        self.result = self._query.execute()

    def _perform_bulk_action(self) -> None:
        for msg in self.result:
            self._action(msg, *self._args, **self._kwargs)


class BulkAction:
    def __init__(self, query: Query) -> None:
        self._query = query



import pytest


def test_BlobStorage___init__():
    assert True


def test_BlobStorage_authenticate():
    assert True


def test_Manager___init__():
    assert True


def test_Office___init__():
    assert True


def test_Office___repr__():
    assert True


def test_Office_authenticate():
    assert True


def test_Office_request_token():
    assert True


def test_Outlook___init__():
    assert True


def test_People___init__():
    assert True



import pytest


def test_BlobContainerNameSpace___init__():
    assert True


def test_BlobContainer___bool__():
    assert True


def test_BlobContainer___getitem__():
    assert True


def test_BlobContainer___init__():
    assert True


def test_BlobContainer___iter__():
    assert True


def test_BlobContainer___len__():
    assert True


def test_BlobContainer___next__():
    assert True


def test_BlobContainer___repr__():
    assert True


def test_BlobContainer___str__():
    assert True


def test_BlobContainer_download_blob_to():
    assert True


def test_BlobContainer_upload_blob_from():
    assert True


def test_Blob___init__():
    assert True


def test_Blob___repr__():
    assert True


def test_Blob_download_to():
    assert True



from __future__ import annotations

from typing import Any, Union, TYPE_CHECKING
import os

from maybe import Maybe

if TYPE_CHECKING:
    from .message import Message
    from .folder import Folder
    from .contact import Contact


class FluentMessage:
    def __init__(self, message: Message = None, parent: Union[Folder, Contact] = None) -> None:
        self.message = Maybe(message).else_(Maybe(parent).new_message().else_(None))
        self.office = Maybe(message).office.else_(Maybe(parent).office.else_(None))
        self._signing = False
        self._temp_body: str = None
        self.message.sender.address = self.office.address

    def from_(self, address: str) -> FluentMessage:
        self.message.sender.address = address
        return self

    def to(self, contacts: Any) -> FluentMessage:
        if isinstance(contacts, (tuple, set, list)):
            ret = [contact if not hasattr(contact, "main_email") else contact.main_email for contact in contacts]
        else:
            ret = contacts if not hasattr(contacts, "main_email") else contacts.main_email

        self.message.to.add(ret)
        return self

    def subject(self, subject: str) -> FluentMessage:
        self.message.subject = subject
        return self

    def body(self, body: str) -> FluentMessage:
        self._temp_body = body.replace("\n", "<br>").replace("\t", "&nbsp;"*4)
        return self

    def attach(self, attachments: Any) -> FluentMessage:
        self.message.attachments.add([os.fspath(attachment) for attachment in attachments] if isinstance(attachments, (list, tuple, set)) else os.fspath(attachments))
        return self

    def sign(self, signing: bool = True) -> FluentMessage:
        self._signing = signing
        return self

    def send(self) -> bool:
        if self._temp_body is not None:
            self.message.body = f"{self._temp_body}<br><br>{self.office.outlook.signature}" if self._signing else self._temp_body

        return self.message.send()
